@@ -24,6 +24,8 @@
 
 #include "grd-context.h"
 
+#include "grd-egl-thread.h"
+
 #include "grd-dbus-remote-desktop.h"
 #include "grd-dbus-screen-cast.h"
 
@@ -40,14 +42,14 @@ struct _GrdContext
   GrdDBusRemoteDesktop *remote_desktop_proxy;
   GrdDBusScreenCast *screen_cast_proxy;
 
-  GrdSettings *settings;
+  GrdEglThread *egl_thread;
 
-  GList *sessions;
+  GrdSettings *settings;
 
   GrdDebugFlags debug_flags;
 };
 
-G_DEFINE_TYPE (GrdContext, grd_context, G_TYPE_OBJECT);
+G_DEFINE_TYPE (GrdContext, grd_context, G_TYPE_OBJECT)
 
 GrdDBusRemoteDesktop *
 grd_context_get_remote_desktop_proxy (GrdContext *context)
@@ -83,34 +85,18 @@ grd_context_get_main_context (GrdContext *context)
   return context->main_context;
 }
 
-static void
-on_session_stopped (GrdSession *session,
-                    GrdContext *context)
-{
-  context->sessions = g_list_remove (context->sessions, session);
-}
-
-void
-grd_context_add_session (GrdContext *context,
-                         GrdSession *session)
-{
-  context->sessions = g_list_append (context->sessions, session);
-  g_signal_connect (session, "stopped",
-                    G_CALLBACK (on_session_stopped), context);
-}
-
-GList *
-grd_context_get_sessions (GrdContext *context)
-{
-  return context->sessions;
-}
-
 GrdSettings *
 grd_context_get_settings (GrdContext *context)
 {
   return context->settings;
 }
 
+GrdEglThread *
+grd_context_get_egl_thread (GrdContext *context)
+{
+  return context->egl_thread;
+}
+
 GrdDebugFlags
 grd_context_get_debug_flags (GrdContext *context)
 {
@@ -139,6 +125,7 @@ grd_context_finalize (GObject *object)
 
   g_clear_object (&context->remote_desktop_proxy);
   g_clear_object (&context->screen_cast_proxy);
+  g_clear_pointer (&context->egl_thread, grd_egl_thread_free);
   g_clear_object (&context->settings);
 
   G_OBJECT_CLASS (grd_context_parent_class)->finalize (object);
@@ -147,11 +134,17 @@ grd_context_finalize (GObject *object)
 static void
 grd_context_init (GrdContext *context)
 {
+  g_autoptr (GError) error = NULL;
+
   context->main_context = g_main_context_default ();
 
   init_debug_flags (context);
 
   context->settings = g_object_new (GRD_TYPE_SETTINGS, NULL);
+
+  context->egl_thread = grd_egl_thread_new (&error);
+  if (!context->egl_thread)
+    g_debug ("Failed to create EGL thread: %s", error->message);
 }
 
 static void
