@@ -37,7 +37,7 @@ typedef enum _GrdDebugFlags
 } GrdDebugFlags;
 
 #define GRD_TYPE_CONTEXT (grd_context_get_type ())
-G_DECLARE_FINAL_TYPE (GrdContext, grd_context, GRD, CONTEXT, GObject);
+G_DECLARE_FINAL_TYPE (GrdContext, grd_context, GRD, CONTEXT, GObject)
 
 GrdDBusRemoteDesktop * grd_context_get_remote_desktop_proxy (GrdContext *context);
 
@@ -53,13 +53,10 @@ GrdPipeWireStreamMonitor *grd_context_get_pipewire_stream_monitor (GrdContext *c
 
 GMainContext *grd_context_get_main_context (GrdContext *context);
 
-void grd_context_add_session (GrdContext *context,
-                              GrdSession *session);
-
-GList * grd_context_get_sessions (GrdContext *context);
-
 GrdSettings * grd_context_get_settings (GrdContext *context);
 
+GrdEglThread * grd_context_get_egl_thread (GrdContext *context);
+
 GrdDebugFlags grd_context_get_debug_flags (GrdContext *context);
 
 #endif /* GRD_CONTEXT_H */
