@@ -25,6 +25,8 @@
 #include <gio/gio.h>
 #include <string.h>
 
+#include "grd-schemas.h"
+
 #define GRD_RDP_SCHEMA_ID "org.gnome.desktop.remote-desktop.rdp"
 #define GRD_VNC_SCHEMA_ID "org.gnome.desktop.remote-desktop.vnc"
 #define GRD_RDP_SERVER_PORT 3389
@@ -32,9 +34,12 @@
 
 enum
 {
+  RDP_ENABLED_CHANGED,
+  RDP_SCREEN_SHARE_MODE_CHANGED,
   RDP_SERVER_CERT_CHANGED,
   RDP_SERVER_KEY_CHANGED,
   RDP_VIEW_ONLY_CHANGED,
+  VNC_ENABLED_CHANGED,
   VNC_VIEW_ONLY_CHANGED,
   VNC_AUTH_METHOD_CHANGED,
   VNC_ENCRYPTION_CHANGED,
@@ -50,6 +55,9 @@ struct _GrdSettings
 
   struct {
     GSettings *settings;
+
+    gboolean is_enabled;
+    GrdRdpScreenShareMode screen_share_mode;
     char *server_cert;
     char *server_key;
     gboolean view_only;
@@ -57,6 +65,7 @@ struct _GrdSettings
   } rdp;
   struct {
     GSettings *settings;
+    gboolean is_enabled;
     gboolean view_only;
     GrdVncAuthMethod auth_method;
     int port;
@@ -65,36 +74,6 @@ struct _GrdSettings
 
 G_DEFINE_TYPE (GrdSettings, grd_settings, G_TYPE_OBJECT)
 
-const SecretSchema *
-grd_rdp_credentials_get_schema (void)
-{
-  static const SecretSchema grd_rdp_credentials_schema = {
-    .name = "org.gnome.RemoteDesktop.RdpCredentials",
-    .flags = SECRET_SCHEMA_NONE,
-    .attributes = {
-      { "credentials", SECRET_SCHEMA_ATTRIBUTE_STRING },
-      { "NULL", 0 },
-    },
-  };
-
-  return &grd_rdp_credentials_schema;
-}
-
-const SecretSchema *
-grd_vnc_password_get_schema (void)
-{
-  static const SecretSchema grd_vnc_password_schema = {
-    .name = "org.gnome.RemoteDesktop.VncPassword",
-    .flags = SECRET_SCHEMA_NONE,
-    .attributes = {
-      { "password", SECRET_SCHEMA_ATTRIBUTE_STRING },
-      { "NULL", 0 },
-    },
-  };
-
-  return &grd_vnc_password_schema;
-}
-
 int
 grd_settings_get_rdp_port (GrdSettings *settings)
 {
@@ -121,6 +100,12 @@ grd_settings_override_vnc_port (GrdSettings *settings,
   settings->vnc.port = port;
 }
 
+GrdRdpScreenShareMode
+grd_settings_get_screen_share_mode (GrdSettings *settings)
+{
+  return settings->rdp.screen_share_mode;
+}
+
 char *
 grd_settings_get_rdp_server_cert (GrdSettings *settings)
 {
@@ -137,14 +122,14 @@ char *
 grd_settings_get_rdp_username (GrdSettings  *settings,
                                GError      **error)
 {
-  const char *test_password_override;
+  const char *test_username_override;
   g_autofree char *credentials_string = NULL;
   g_autoptr (GVariant) credentials = NULL;
   char *username = NULL;
 
-  test_password_override = g_getenv ("GNOME_REMOTE_DESKTOP_TEST_RDP_PASSWORD");
-  if (test_password_override)
-    return g_strdup ("TEST");
+  test_username_override = g_getenv ("GNOME_REMOTE_DESKTOP_TEST_RDP_USERNAME");
+  if (test_username_override)
+    return g_strdup (test_username_override);
 
   credentials_string = secret_password_lookup_sync (GRD_RDP_CREDENTIALS_SCHEMA,
                                                     NULL, error,
@@ -239,6 +224,18 @@ grd_settings_get_vnc_view_only (GrdSettings *settings)
   return settings->vnc.view_only;
 }
 
+gboolean
+grd_settings_is_rdp_enabled (GrdSettings *settings)
+{
+  return settings->rdp.is_enabled;
+}
+
+gboolean
+grd_settings_is_vnc_enabled (GrdSettings *settings)
+{
+  return settings->vnc.is_enabled;
+}
+
 GrdVncAuthMethod
 grd_settings_get_vnc_auth_method (GrdSettings *settings)
 {
@@ -248,6 +245,20 @@ grd_settings_get_vnc_auth_method (GrdSettings *settings)
     return settings->vnc.auth_method;
 }
 
+static void
+update_screen_share_mode (GrdSettings *settings)
+{
+  settings->rdp.screen_share_mode =
+    g_settings_get_enum (settings->rdp.settings, "screen-share-mode");
+}
+
+static void
+update_rdp_enabled (GrdSettings *settings)
+{
+  settings->rdp.is_enabled = g_settings_get_boolean (settings->rdp.settings,
+                                                     "enable");
+}
+
 static void
 update_rdp_tls_cert (GrdSettings *settings)
 {
@@ -271,6 +282,13 @@ update_rdp_view_only (GrdSettings *settings)
                                                     "view-only");
 }
 
+static void
+update_vnc_enabled (GrdSettings *settings)
+{
+  settings->vnc.is_enabled = g_settings_get_boolean (settings->vnc.settings,
+                                                     "enable");
+}
+
 static void
 update_vnc_view_only (GrdSettings *settings)
 {
@@ -290,7 +308,17 @@ on_rdp_settings_changed (GSettings   *rdp_settings,
                          const char  *key,
                          GrdSettings *settings)
 {
-  if (strcmp (key, "tls-cert") == 0)
+  if (strcmp (key, "enable") == 0)
+    {
+      update_rdp_enabled (settings);
+      g_signal_emit (settings, signals[RDP_ENABLED_CHANGED], 0);
+    }
+  else if (strcmp (key, "screen-share-mode") == 0)
+    {
+      update_screen_share_mode (settings);
+      g_signal_emit (settings, signals[RDP_SCREEN_SHARE_MODE_CHANGED], 0);
+    }
+  else if (strcmp (key, "tls-cert") == 0)
     {
       update_rdp_tls_cert (settings);
       g_signal_emit (settings, signals[RDP_SERVER_CERT_CHANGED], 0);
@@ -312,7 +340,12 @@ on_vnc_settings_changed (GSettings   *vnc_settings,
                          const char  *key,
                          GrdSettings *settings)
 {
-  if (strcmp (key, "view-only") == 0)
+  if (strcmp (key, "enable") == 0)
+    {
+      update_vnc_enabled (settings);
+      g_signal_emit (settings, signals[VNC_ENABLED_CHANGED], 0);
+    }
+  else if (strcmp (key, "view-only") == 0)
     {
       update_vnc_view_only (settings);
       g_signal_emit (settings, signals[VNC_VIEW_ONLY_CHANGED], 0);
@@ -348,9 +381,12 @@ grd_settings_init (GrdSettings *settings)
   g_signal_connect (settings->vnc.settings, "changed",
                     G_CALLBACK (on_vnc_settings_changed), settings);
 
+  update_rdp_enabled (settings);
+  update_screen_share_mode (settings);
   update_rdp_tls_cert (settings);
   update_rdp_tls_key (settings);
   update_rdp_view_only (settings);
+  update_vnc_enabled (settings);
   update_vnc_view_only (settings);
   update_vnc_auth_method (settings);
 
@@ -365,6 +401,20 @@ grd_settings_class_init (GrdSettingsClass *klass)
 
   object_class->finalize = grd_settings_finalize;
 
+  signals[RDP_ENABLED_CHANGED] =
+    g_signal_new ("rdp-enabled-changed",
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_LAST,
+                  0,
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE, 0);
+  signals[RDP_SCREEN_SHARE_MODE_CHANGED] =
+    g_signal_new ("rdp-screen-share-mode-changed",
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_LAST,
+                  0,
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE, 0);
   signals[RDP_SERVER_CERT_CHANGED] =
     g_signal_new ("rdp-tls-cert-changed",
                   G_TYPE_FROM_CLASS (klass),
@@ -386,6 +436,13 @@ grd_settings_class_init (GrdSettingsClass *klass)
                   0,
                   NULL, NULL, NULL,
                   G_TYPE_NONE, 0);
+  signals[VNC_ENABLED_CHANGED] =
+    g_signal_new ("vnc-enabled-changed",
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_LAST,
+                  0,
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE, 0);
   signals[VNC_VIEW_ONLY_CHANGED] =
     g_signal_new ("vnc-view-only-changed",
                   G_TYPE_FROM_CLASS (klass),
