@@ -30,10 +30,9 @@
 G_DECLARE_FINAL_TYPE (GrdSettings, grd_settings,
                       GRD, SETTINGS, GObject)
 
-const SecretSchema * cc_grd_rdp_credentials_get_schema (void);
-const SecretSchema * cc_grd_vnc_password_get_schema (void);
-#define GRD_RDP_CREDENTIALS_SCHEMA grd_rdp_credentials_get_schema ()
-#define GRD_VNC_PASSWORD_SCHEMA grd_vnc_password_get_schema ()
+gboolean grd_settings_is_rdp_enabled (GrdSettings *settings);
+
+gboolean grd_settings_is_vnc_enabled (GrdSettings *settings);
 
 int grd_settings_get_rdp_port (GrdSettings *settings);
 
@@ -45,6 +44,8 @@ void grd_settings_override_rdp_port (GrdSettings *settings,
 void grd_settings_override_vnc_port (GrdSettings *settings,
                                      int          port);
 
+GrdRdpScreenShareMode grd_settings_get_screen_share_mode (GrdSettings *settings);
+
 char * grd_settings_get_rdp_server_cert (GrdSettings *settings);
 
 char * grd_settings_get_rdp_server_key (GrdSettings *settings);
